(*
 *  beware IRC services, authaccount.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  modifications by peter green
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)
unit accountcore;
{
possible load accounts code:

[A.account]
- currentaccount := nil

name=
- if accounts exists, make it current, otherwise, create new account, add it to list, and make it current

password=
- crypted password

level=
- change level of current account

mask=mask1 mask2 mask3
- set mask list of account to these masks, remove existing masks
maskmore=mask1 mask2 mask3
- add masks to list, check for already existing masks


- an account without password is disabled (cant login)
set with MODOPER SETPASS idnick <newpass>

if the account is mentioned in the cfg file, set a flag. at the end of loading,
delete all accounts which dont have the flag set
}




interface

uses
  blinklist , bstuff , buser , objbot , bservicemngr , bconsts , pgtypes , bvaliddef;

const
  idnicklen=16;

type
  taccountcore=class; // for those less familiar with theese things this is
                      // a forward declaration to allow taccount to referce
                      // taccountcore
  taccount=class(tlinklist)
    name:string;
    password:string;
    requestedpass:string; {extra, temp, password}
    level:integer;
    mask:tstringlinklist;

    user:tplinklist;
    {who is logged in to this account, linked list of pointers, nil=none}

    marked:boolean; {internal}
    operonly:boolean;
    email : string; {email set on registration can only be changed by admins}
    hellouser : string; {  "nick!user@host ip" form
                         set on registration with hello command
                         empty string for accounts created by ircops
                         cannot be changed by anyone. used to issue glines
                         when a user creates an account USING someone elses
                         email (this is much safer than trusting the
                         nick!user@host in the bounce/reply}

    lastseents:integer; {for expire purposes}
    lastmailts:integer; {requestpass throttle}
    accountcore : taccountcore;
    // the following 2 fields are only used in mode 2
    idstr        : string;
    nickchangets : integer;
    bottag:array[0..maxserviceunit] of pointer; {4 bytes for each service unit}
    function addmask(s:string):boolean;
    function delmask(s:string):boolean;
    function adduser(us:tuser):boolean;
    function removeuser(us:tuser):boolean;
    procedure clearmasks;
    constructor create;
    destructor destroy; override;
  end;
  taccountcore=class(tobjbot)
  public
    accountlist:taccount;

    mode : byte;
    // 0: use ac tokens to recognise returning users
    // 1: use timestamps to recognise returning users
    // 2: don't recognise returning users
    accountsfile : string;

    function findaccount(const s:string):taccount;

    {add account. only to be used from command handler or similar}
    function addaccount(const name:string;level:integer):taccount;

//    function validaccountname(const s:string):boolean;
    function saneaccountname(const s:string):boolean;

    function convertaccountname(const s:string):string;

    procedure clearaccount;
    procedure saveaccount(const filename:string);
    procedure loadaccount(const filename:string);

//    procedure updatevalidaccounttable;

    function accountpasswordmatch(p:taccount;password:string):boolean;
    procedure login(us:tuser;account:string);
    procedure deauth(us:tuser);
    function getaccountfromuser(us:tuser):taccount;
    procedure setupsrvunit;override;
    procedure changemode(newmode:byte);
    function levelcontrol(mylevel,otherlevel:integer):boolean;
    procedure onnewaccount(ac:taccount);
    procedure ondestroyaccount(ac:taccount);
    procedure onlogin(us:tuser;ac:taccount);
    procedure ondeauth(us:tuser;ac:taccount);

  end;
  tonnewaccount=procedure(ac:taccount;tag:taddrint);
  tonlogin=procedure(us:tuser;ac:taccount;tag:taddrint);
  taccountUSINGserviceunit=class(tserviceunit)
  public
    accountcore      : taccountcore  ;
    onnewaccount     : tonnewaccount ;
    ondestroyaccount : tonnewaccount ;
    onlogin          : tonlogin      ;
    ondeauth         : tonlogin      ;
  end;

var
  globalaccountcore : taccountcore;
implementation

uses
  bconfig , passcryp , bparse    , bcmds      , sysutils , unitwritetxt ,
  bcrash  , btime    , bservunit , pgdebugout , bchannel ;



procedure taccountcore.onnewaccount(ac:taccount);
var
  a:integer;
  s:tserviceunit;
begin
  //ipcheck_remoteclient(ac);
  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then
     if s is taccountUSINGserviceunit then
      if taccountUSINGserviceunit(s).accountcore = self then
       if @taccountUSINGserviceunit(s).onnewaccount <> nil then
        taccountUSINGserviceunit(s).onnewaccount(ac,s.tag);
  end;
end;

procedure taccountcore.ondestroyaccount(ac:taccount);
var
  a:integer;
  s:tserviceunit;
begin
  //ipcheck_remoteclient(ac);
  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then
     if s is taccountUSINGserviceunit then
      if taccountUSINGserviceunit(s).accountcore = self then
       if @taccountUSINGserviceunit(s).ondestroyaccount <> nil then
        taccountUSINGserviceunit(s).ondestroyaccount(ac,s.tag);
  end;
end;

procedure taccountcore.onlogin(us:tuser;ac:taccount);
var
  a:integer;
  s:tserviceunit;
begin
  //ipcheck_remoteclient(ac);
  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then
     if s is taccountUSINGserviceunit then
      if taccountUSINGserviceunit(s).accountcore = self then
       if @taccountUSINGserviceunit(s).onlogin <> nil then
        taccountUSINGserviceunit(s).onlogin(us,ac,s.tag);
  end;
end;

procedure taccountcore.ondeauth(us:tuser;ac:taccount);
var
  a:integer;
  s:tserviceunit;
begin
  //ipcheck_remoteclient(ac);
  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then
     if s is taccountUSINGserviceunit then
      if taccountUSINGserviceunit(s).accountcore = self then
       if @taccountUSINGserviceunit(s).ondeauth <> nil then
        taccountUSINGserviceunit(s).ondeauth(us,ac,s.tag);
  end;
end;


function taccountcore.saneaccountname(const s:string):boolean;
var
  a:integer;
begin
  result := false;
  if (s = '') or (length(s) > 63) then exit;
  for a := 1 to length(s) do if s[a] <= #32 then exit;
  result := true;
end;

{
}

{
centralized code to convert (or not) to uppercase
}
function taccountcore.convertaccountname(const s:string):string;
begin
  result := ircupper(s);
end;
{
IDnick's are case tolerant here
i may later make hashtable code for this,
but there arent thousands of accounts anyway, so this will do for now
}
function taccountcore.findaccount(const s:string):taccount;
var
  p:taccount;
begin
  p := accountlist;
  result := nil;
  while p <> nil do begin
    if strcompup(p.name,s) then begin
      result := p;
      exit
    end;
    p := taccount(p.next);
  end;
end;


function taccountcore.accountpasswordmatch(p:taccount;password:string):boolean;
begin
  result := ((p.password <> '') and passmatch(password,p.password))
  or ((p.requestedpass <> '') and passmatch(password,p.requestedpass));
end;


function taccountcore.addaccount(const name:string;level:integer):taccount;
var
  p:taccount;
begin
  result := nil;
  if findaccount(name) <> nil then exit; {aready exists}
  p := taccount.create;
  p.accountcore := self;
  linklistadd(tlinklist(accountlist),tlinklist(p));
  p.name := name;
  p.level := level;
  onnewaccount(p);
  result := p;
end;

function taccount.addmask(s:string):boolean;
var
  s2:string;
  sl:tstringlinklist;

begin
  result := false;

  s := cookmask(s);
  {process mask (like ban), must also check here if mask is valid}

  s2 := ircupper(s);
  sl := self.mask;
  while sl <> nil do begin
    if ircupper(sl.s) = s2 then exit; {mask already exists}
    sl := tstringlinklist(sl.next);
  end;

  sl := tstringlinklist.create;
  sl.s := s;
  linklistadd(tlinklist(self.mask),tlinklist(sl));
  result := true;
end;

function taccount.delmask(s:string):boolean;
var
  s2:string;
  sl,sl2:tstringlinklist;

begin
  result := false;

  s := cookmask(s);
  {process mask (like ban), must also check here if mask is valid}

  s2 := ircupper(s);
  sl2 := nil;
  sl := self.mask;
  while sl <> nil do begin
    if ircupper(sl.s) = s2 then begin
      sl2 := sl;
      break;
    end;
    sl := tstringlinklist(sl.next);
  end;

  if sl2 = nil then exit; {mask doesn't exist}
  linklistdel(tlinklist(self.mask),tlinklist(sl2));
  sl2.destroy;
  result := true;
end;

constructor taccount.create;
begin
  inherited create;
  {}
end;

destructor taccount.destroy;
begin
  accountcore.ondestroyaccount(self);
  while self.user <> nil do accountcore.deauth(tuser(self.user.p));

  {if self.user <> nil then authbot.deauth(self.user);}

  self.clearmasks;
  inherited destroy;
end;

function Taccount.adduser(us:tuser):boolean;
var
  p:tplinklist;
begin
  result := false;
  {search user already in list}
  p := self.user;
  while p <> nil do begin
    if p.p = us then exit;
    p := tplinklist(p.next);
  end;
  result := true;
  p := tplinklist.create;
  p.p := us;
  linklistadd(tlinklist(self.user),tlinklist(p));
end;

function Taccount.removeuser(us:tuser):boolean;
var
  p:tplinklist;
begin
  result := true;
  {search user in list}
  p := tplinklist(self.user);
  while p <> nil do begin
    if p.p = us then begin
      linklistdel(tlinklist(self.user),tlinklist(p));
      p.destroy;
      exit;
    end;
    p := tplinklist(p.next);
  end;
  result := false;
end;

procedure taccount.clearmasks;
var
  sl,sl2:tstringlinklist;
begin
  {destroy all masks}
  sl := self.mask;
  while sl <> nil do begin
    sl2 := tstringlinklist(sl.next);
    linklistdel(tlinklist(self.mask),tlinklist(sl));
    sl.destroy;
    sl := sl2;
  end;
end;

{free A oper account list (to be done on shutdown)}
procedure taccountcore.clearaccount;
var
  p,p2:taccount;
begin
  p := accountlist;
  while p <> nil do begin
    p2 := taccount(p.next);
    linklistdel(tlinklist(accountlist),tlinklist(p));
    p.destroy;
    p := p2;
  end;
end;

var
  currentaccount:taccount;
  currentaccountcore : taccountcore;
procedure loadaccountproc(sv,sn:string);
var
  newaccount : boolean;
begin
 with currentaccountcore do begin
  if sv = 'NAME' then begin
    currentaccount := nil;
    if not saneaccountname(convertaccountname(sn)) then exit;

    currentaccount := findaccount(sn);
    if currentaccount = nil then begin
      currentaccount := taccount.create;
      currentaccount.accountcore := currentaccountcore;
      linklistadd(tlinklist(accountlist),tlinklist(currentaccount));
      newaccount := true;
    end else begin
      currentaccount.clearmasks;
      newaccount := false;
    end;
    currentaccount.name := sn; {if the account already exists, this allows changing the case of the name}
    currentaccount.marked := true;
    if newaccount then onnewaccount(currentaccount);
  end else if currentaccount <> nil then begin
    if sv = 'LEVEL' then begin
      currentaccount.level := strtointdef(sn,0);
    end else if sv = 'MASK' then begin
      sn := cookmask(sn);
      if validmask(sn) then currentaccount.addmask(sn);
    end else if sv = 'PASSWORD' then begin
      if iscryptedlength(length(sn)) or (sn = '') then
      currentaccount.password := sn
      else
      currentaccount.password := passnewcrypt(sn);
    end else if sv = 'OPERONLY' then begin
      currentaccount.operonly := sn = '1';
    end else if sv = 'EMAIL' then begin
      currentaccount.email := sn;
    end else if sv = 'HELLOUSER' then begin
      currentaccount.hellouser := sn;
    end else if sv = 'LASTSEEN' then begin
      currentaccount.lastseents := strtointdef(sn,0);
    end else if sv = 'LASTMAIL' then begin
      currentaccount.lastmailts := strtointdef(sn,0);
    end;

  end;
 end;
end;


procedure taccountcore.loadaccount(const filename:string);
var
  p,p2:taccount;

begin
  currentaccountcore := self;
  //writeln('loadAaccount called');
  {don't clear list}

  {un-mark all accounts}
  p := accountlist;
  while p <> nil do begin
    p.marked := false;
    p := taccount(p.next);
  end;

  currentaccount := nil;
  readcfg(filename,@loadaccountproc);
  currentaccount := nil;

  p := accountlist;
  while p <> nil do begin
    p2 := taccount(p.next);
    if not p.marked then begin
      linklistdel(tlinklist(Accountlist),tlinklist(p));
      p.destroy;
    end;
    p := p2;
  end;
end;

procedure taccountcore.saveaccount(const filename:string);
var
  t:textfile;
  sl:tstringlinklist;
  p:taccount;
begin
  if not openforwrite(t,filename) then exit;
  p := tAccount(Accountlist);
  while p <> nil do begin
    {}
    writeln(t,'[account]');
    writeln(t,'name='+p.name);
    writeln(t,'level='+inttostr(p.level));
    writeln(t,'password='+p.password);
    writeln(t,'operonly='+inttostr(ord(p.operonly)));
    writeln(t,'email='+p.email);
    writeln(t,'hellouser='+p.hellouser);
    writeln(t,'lastseen='+inttostr(p.lastseents));
    writeln(t,'lastmail='+inttostr(p.lastmailts));
    sl := p.mask;
    while sl <> nil do begin
      writeln(t,'mask='+sl.s);
      sl := tstringlinklist(sl.next);
    end;
    p := taccount(p.next);
    writeln(t);
  end;
  closefile(t);
end;

procedure taccountcore.deauth(us:tuser);
var
  t:taccount;
begin
  t := getaccountfromuser(us);
  if t = nil then exit;
  ondeauth(us,t);
  assert(t.removeuser(us),'Auth.deauth: user not in account','');
  us.bottag[srvunit.num] := nil;
  t.lastseents := unixtime;

  if up then if not flag_isset(us.flags,userflag_destroying) then sendkill(us,me.name,'deauth');

  us.bottag[srvunit.num] := nil;
end;

function taccountcore.getaccountfromuser(us:tuser):taccount;
begin
  //debugout(inttostr(taddrint(us)));
  //debugout(inttostr(taddrint(srvunit)));
  result := us.bottag[srvunit.num];

end;
procedure onnickchange(us:tuser;tag:taddrint);
var
  P : taccount;
begin
  with taccountcore(tag) do begin
    if mode=1 then begin
      p := getaccountfromuser(us);
      if assigned(p) then begin
        p.idstr := us.idstr;
        p.nickchangets := us.ts;
      end;

    end;
  end;
end;
procedure taccountcore.login(us:tuser;account:string);
var
  P : taccount;
begin

  if (mode=0)and(us.account = '') then begin
    sendparse(me.idstr+' '+tok_account+' '+us.idstr+' '+account);
  end else begin
    p := findaccount(account);
    if p = nil then begin
      p := addaccount(account,0);
      //logaction(us,'account does not exist: '+account);
    end;

    p.lastseents := unixtime;

    us.bottag[srvunit.num] := p;
    {p.user := us;}
    p.adduser(us);
    onnickchange(us,srvunit.tag);
    onlogin(us,p);
  end;
end;


procedure onshutdown(tag:taddrint);
var
  a           : integer;
begin
  with taccountcore(tag) do begin
    if not up then exit;
    up := false;
    for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
      if @serviceunitarray[a].onshutdown <> nil then
        if serviceunitarray[a] is taccountUSINGserviceunit then
          if taccountUSINGserviceunit(serviceunitarray[a]).accountcore=taccountcore(tag) then
            serviceunitarray[a].onshutdown(serviceunitarray[a].tag);
    saveaccount(accountsfile);
    clearaccount;
    //bot_quit(mybot,shutdownreason);
    //mybot := nil;
  end;


end;

procedure onaccount(sptr:tuser;us:tuser;tag:taddrint);
var
  ac : taccount;
begin
  with taccountcore(tag) do begin
    if up and (mode=0) then begin
      login(us,us.account);
      ac := getaccountfromuser(us);
      if ac.level = 0 then ac.level := 1;
    end;
  end;
end;

function expandidstr(idstr:string) : string;

begin
  result := idstr;
  case length(idstr) of
    1: begin
      // S becomes AS
      result := 'A'+idstr;
    end;
    3: begin
      // SCC becomes ASACC
      setlength(result,5);
      result[1] := 'A';
      result[2] := idstr[1];
      result[3] := 'A';
      result[4] := idstr[2];
      result[5] := idstr[3];

    end;
  end;
end;

procedure onnewuser(us:tuser;tag:taddrint);
var
  p:taccount;
begin
  with taccountcore(tag) do begin

    if up and (mode=1) then begin
      p := accountlist;
      //result := nil;
      while p <> nil do begin
        if expandidstr(us.idstr)=expandidstr(p.idstr) then begin
          if us.ts=p.nickchangets then begin
            login(us,p.name);
          end else begin
            requestsignonts(us);

          end;

        end;
        p := taccount(p.next);

      end;
    end;

  end;
end;

procedure ondestroyuser(us:tuser;tag:taddrint);
begin
  with taccountcore(tag) do begin
    deauth(us);
    {}
  end;
end;

procedure onstartup(tag:taddrint);
var
  currentuser : tuser;
begin
  with taccountcore(tag) do begin
    if up then exit;
    up := true;
    loadaccount(accountsfile);
    //mybot := addbot(srvunit,nick,userid,host,realname,0);
    //prefix := sendprefixstr(mybot);

    //startup login processing

    currentuser := tuser(globaluserlist);

    while assigned(currentuser) do begin
      case mode of
        0: begin
          if currentuser.account <> '' then onaccount(me,currentuser,srvunit.tag);
        end;
        1: begin
          onnewuser(currentuser,srvunit.tag);
        end;
      end;
      currentuser := tuser(currentuser.next);

    end;
  end;
end;

procedure onsignonts(sptr:tuser;us:tuser;tag:taddrint);
var
  p:taccount;

begin
  with taccountcore(tag) do begin

    if up and (mode=1) then begin
      p := accountlist;
      //result := nil;
      while p <> nil do begin
        if expandidstr(us.idstr)=expandidstr(p.idstr) then begin
          //first check is a sanity check
          if (us.signonts > globaldefaultts)and(us.signonts<p.nickchangets) then begin
            login(us,p.name);
          end;

        end;
        p := taccount(p.next);

      end;
    end;

  end;
end;

procedure onsave(tag:taddrint;saveall:boolean);
begin
  with taccountcore(tag) do begin
    if up then begin
      saveaccount(accountsfile);
    end;
  end;
end;
procedure taccountcore.setupsrvunit;
begin
  inherited setupsrvunit;
  srvunit.onstartup     := onstartup     ;
  srvunit.onshutdown    := onshutdown    ;
  srvunit.onaccount     := onaccount     ;
  srvunit.onnewuser     := onnewuser     ;
  srvunit.ondestroyuser := ondestroyuser ;
  srvunit.onnickchange  := onnickchange  ;
  srvunit.onsignonts    := onsignonts    ;
  srvunit.onsave        := onsave        ;
end;

procedure taccountcore.changemode(newmode:byte);
begin
  if up then begin
    onshutdown(srvunit.tag);
    mode:=newmode;
    onstartup(srvunit.tag);
  end else begin
    mode := newmode;
  end;
end;

function taccountcore.levelcontrol(mylevel,otherlevel:integer):boolean;
begin
  if mylevel >= maxlevel then begin
    result := mylevel >= otherlevel;
  end else begin
    result := mylevel > otherlevel;
  end;
end;


initialization
  globalaccountcore := taccountcore.create;
  with globalaccountcore do begin
    accountsfile := 'accounts.ini';
    if (not fileexists('accounts.ini')) and fileexists('auth.accounts.ini') then begin
      renamefile('auth.accounts.ini','accounts.ini');
    end;
    //mode := 1;
  end;
end.
