(*
 *  bserv, Internet Relay Chat services,
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  some modifications by peter green
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)
program bserv;

{$include bserv.inc}

uses
  {$include bots.inc}
  {$ifndef nowinnt}
    bwinnt,
  {$endif}
  {$ifndef nosignal}
  bsignal,
  {$endif}
  {$ifndef win32}
    {$ifdef VER1_0}
      linux,
    {$else}
      baseunix,unix,
    {$endif}
    unixsignals,unitfork,
  {$endif}
  bservunit,
  bconfig,
  btime,
  bstuff,
  sysutils,
  pgdebugout,
  replymanager;

  {$ifndef noicon}
    {$R *.res}
  {$endif}

  {names of units, do not remove -beware
  lcore lsocket wcore wsocket authbot bworldbot accountuseingbot objbot
  connect4bot invkickbot authaccount bconsts bparse b_burst buser bchannel
  bservicemngr bcrash connect4bot invkickbot boptiontable bsetcommand channelbot
  channeldb b_burst bipcheck unitbanmask

  }
{$include unixstuff.inc}

{$ifndef win32}
const pidfile='bserv.pid';
{$endif}

var
  counter    : integer;
  p          : string;
  foreground : boolean;
  nopid      : boolean;
begin
  

  counter := 1;
  while counter <= paramcount do begin
    p := ircupper(paramstr(counter));
    if p = 'DEBUGC' then begin
      debugmode := debugmode or DMConsole;
      foreground := true;
    end else if p = 'DEBUGF' then begin
      debugmode := debugmode or DMFILE;
      inc(counter);
      debugfile := paramstr(counter);
    end else if p = 'FOREGROUND' then begin
      foreground := true;
    end else if p = 'NOPID' then begin
      nopid := true;
    end else ;
    inc(counter);
  end;
  debugout('processed command line params');
  {load core config files}
  getpath;
  randomize;
  debugout('about to call btime.init');
  btime.init;
  debugout('btime.init done');

  bconfig.init;
  debugout('bconfig.init done');
  {$ifndef nosignal}
    if ircupper(paramstr(1)) = 'SIGNAL' then begin
      sendsignal2;
      halt;
    end;
  {$endif}
  {$ifndef nowinnt}
    if ircupper(paramstr(1)) = 'INSTALL' then begin
      installservice;
      opt.runasservice := true;
      writecfg;
      halt;
    end;

    if ircupper(paramstr(1)) = 'UNINSTALL' then begin
      uninstallservice;
      opt.runasservice := false;
      writecfg;
      halt;
    end;
    if opt.runasservice then begin
      runservice;
      halt;
    end;
  {$endif}

  {$ifdef unix}
    if not nopid then begin
      debugout('about to call checkpid');

      if not checkpid(pidfile) then begin

        debugout('halting because of checkpid');
        halt;
        {pid file exists, program still runs or did not cleanly stop}
      end;
      debugout('checkpid done');
      unixsignals.init;
    end;
  {$endif}

  debugout('about to call initapplication');
  initapplication;
  debugout('initapplication complete');

  {$ifdef unix}
  if not foreground then dofork('bserv');
  if not nopid then writepid;
  {$endif}
  runapplication;
  cleanupapplication;

  {$ifdef unix}
  deletepid;
  {$endif}
end.
