(*
 *  beware IRC services, bserver.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

unit bserver;

interface

uses
  {$ifdef win32}wcore,{$else}lcore,{$endif}
  buser,blinklist,bstuff,bconsts;

type
  pointerarray=array[0..0] of pointer;
  ppointerarray=^pointerarray;

var
  p10server:array[0..SSmask] of tserver;

  {all Tservers (including me.server)}
  globalserverlist:tlinklist;

  p10currentslot:integer;
  bootts:integer;
  starttime:integer;
  safesend:boolean; {if set, don't do sendQ exceeded}
  serverisrunning:boolean;
  netsplitquitreason : string;

function addserver(us:tuser;p10num,p10max:integer;parentserver:tserver):tserver;
procedure destroyserver(self:tserver);
procedure init;

{
for "/request <server>" like commands,
this returns the Tuser of the server;
checks servername, nick, and *servermask

if fromclient = false, checks the numeric, otherwise server name mask and nick
}
function getremoteserver(s:string;fromclient:boolean):tuser;

procedure markdestroyingserver(srv:tserver);

implementation

uses
  bparse,bservunit,bchannel,bservicemngr;

function addserver(us:tuser;p10num,p10max:integer;parentserver:tserver):tserver;
begin
  result := tserver.create;
  result.protoversion := 10;
  linklistadd(globalserverlist,tlinklist(result));

  result.us := us;
  us.server := result;

  setflag(us.flags,userflag_isserver);

  result.p10num := p10num;
  p10server[p10num] := result;

  us.idstr := p10inttostr(p10num,SSlen);

  result.p10max := p10max;
  result.parentserver := parentserver;
  getmem(result.p10slots,(p10max+1) shl 2);
  fillchar(result.p10slots^,(p10max+1) shl 2,0);

  inc(count.globalservers);
end;

procedure init;
var
  a:integer;
begin
  globalserverlist := nil;
  for a := 0 to SSmask do p10server[a] := nil;
end;

function getremoteserver(s:string;fromclient:boolean):tuser;
var
  p:tserver;
begin
  result := nil;
  if (s = '') or (s = '*') then begin
    result := me;
    exit;
  end;
  if not fromclient then begin
    result := findnumeric(s);
    exit;
  end;
  s := ircupper(s);

  {try exact match}
  p := tserver(findname(s));
  if p <> nil then begin
    result := tuser(tuser(p).server.us);
    exit;
  end;
  {try maskmatch of servername}
  p := tserver(globalserverlist);
  while p <> nil do begin
    if maskmatchup(s,tuser(p.us).name) then begin
      result := tuser(p.us);
      exit;
    end;
    p := tserver(p.next);
  end;
end;

procedure destroyserver;
var
  p,p2:tlinklist;
  counter:integer;
  uc : tuserchan;
begin
  if serverisrunning then processmessages;
  counter := 0;

  p := globaluserlist;
  while p <> nil do begin
    p2 := p.next;
    if (tuser(p).server = self) and isclient(tuser(p)) then begin
      {don't send a QUIT to other servers, for a user which splits off}
      uc := tuserchan(tuser(p).channel);
      while Assigned(uc) do begin
        onleavechannel(uc.ch,tuser(self.us),tuser(p),lc_quit+lc_netsplit,netsplitquitreason);
        uc := tuserchan(uc.next);
      end;
      tuser(p).destroy;

      inc(counter);
      if counter and 255 = 0 then if serverisrunning then processmessages;
    end;
    p := p2;
  end;

  freemem(self.p10slots);
  p10server[self.p10num] := nil;
  dec(count.globalservers);
  linklistdel(globalserverlist,self);
end;

procedure markdestroyingserver2(srv:tserver);
var
  p:tlinklist;
begin
  if flag_isset(srv.flags,servflag_destroying) then exit;
  setflag(srv.flags,servflag_destroying);
  p := globalserverlist;
  while p <> nil do begin
    if tserver(p).parentserver = srv then markdestroyingserver2(tserver(p));
    p := p.next;
  end;
end;

procedure markdestroyingserver(srv:tserver);
var
  p:tlinklist;
begin
  if flag_isset(srv.flags,servflag_destroying) then exit;
  setflag(srv.flags,servflag_destroying);
  setflag(srv.flags,servflag_destroying_root);
  p := globalserverlist;
  while p <> nil do begin
    if tserver(p).parentserver = srv then markdestroyingserver2(tserver(p));
    p := p.next;
  end;
end;

begin
  {they may already be zero'd but i dont take any risk}
  serverisrunning := false;
  me := nil;
  hostservuser := nil;
  globalserverlist := nil;
end.
