(*
 *  beware IRC services, bservunit.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  some modifications by peter green
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)
unit bservunit;

interface

{$include bserv.inc}

uses
  {$ifdef win32}
    Windows,winsock,wsocket,wcore,
  {$else}
    lsocket,sockets,unixsignals,lcore,
  {$endif}
  SysUtils,Classes,buser,blinklist,bstuff,btime,pgdebugout;


type
  tsc=class
    procedure connected(sender:tobject; error:word);
    procedure closed(sender:tobject; error:word);
    procedure receive(sender:tobject; error:word);
    procedure timehandler(sender:tobject);
  end;

var
  sc:tsc;
  reconnectcountdown:integer;
  wanttoconnect:boolean;
  socketbusy:boolean;
  sock:twsocket;
  shutdownreason:string;

procedure startconnect;
procedure globalshutdown;
procedure startup;


//procedure processmessages;
procedure initapplication;
procedure runapplication;
procedure cleanupapplication;

procedure filelog(filename,line:string);
procedure requestsignonts(us:tuser);
procedure saveall;
const
  botstartfile='startbots.ini';

implementation

uses
  {$ifndef nowinnt}bwinnt,{$endif}
  {$ifndef nosignal}bsignal,{$endif}
  bparse,bconfig,bservicemngr,bserver,bconsts,bcmds,readtxt,bcrash;

var
  recvq:string;

procedure serverdown;
var
  a:integer;
begin
  if me = nil then exit;

  {shutdown bots}
  for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
  if @serviceunitarray[a].onshutdown <> nil then
  serviceunitarray[a].onshutdown(serviceunitarray[a].tag);

  {$ifndef nosignal}
  {close signal port}
  signalstop;
  {$endif}

  if isconnected then begin
    send(me.idstr+' '+TOK_SQUIT+' '+me.name+' 0 :'+shutdownreason);
    sendflush;
    sock.close;
  end;

  {shutdown core}
  writecfg;
  me.destroy;
  me := nil;

  sock.destroy;
  sock := nil;
end;


procedure globalshutdown;
begin
  {$ifndef nowinnt}
  reportstopping;
  {$endif}
  serverdown;
  exitmessageloop;
  {$ifndef nowinnt}
  reportstopped;
  {$endif}
end;


procedure startconnect;
begin
  if isconnected then exit;
  if socketbusy then exit;
  wanttoconnect := true;
  socketbusy := true;

  sock.addr := opt.connectaddr;
  sock.port := opt.connectport;
  sock.proto := 'tcp';
  sock.connect;
end;


procedure tsc.connected;
begin
 try
  debugout('sc.connected called '+inttostr(error));
  if error = 0 then begin
    socketbusy := false;
{    setsockopt(sock.Hsocket,IPPROTO_TCP,TCP_NODELAY,@a,4);}
    recvq := '';
    isconnected := true;
    onconnected;
  end else begin
    isconnected := false;
    socketbusy := false;
  end;
 except
   on e: exception do begin
     send(me.name+' '+TOK_WALLOPS + ' :exception raised dureing sc.connected type '+e.ClassName+' with message '+e.Message+#13#10);
   end;
 end;
end;


procedure Tsc.receive(sender : tobject; error : word);
var
  a,b,c,d:integer;
  s:string;
  goed:boolean;
begin
 try
  recvq := recvq + sock.receivestr;

  goed := false;
  a := 1;
  b := length(recvq);
  d := 0;
  repeat
    if length(recvq)=0 then exit;
    { a squit has resulted in the connection being broken and reset the recvq -pg}
    c := a;
    while (a <= b) and (recvq[a] <> #13) and (recvq[a] <> #10) do inc(a);
    if a <= b then begin
      s := copy(recvq,c,a-c);
      d := a;
      if s <> '' then parsecommand(false,s);
    end else goed := true;
    inc(a);
  until goed;
  if d > 0 then delete(recvq,1,d);
  sendflush;
 except
   on e: exception do begin
     send(me.name+' '+TOK_WALLOPS + ' :exception raised dureing sc.receive type '+e.ClassName+' with message '+e.Message+#13#10);
   end;
 end;
end;

procedure tsc.closed;
begin

  if isconnected then sock.close;
  socketbusy := false;
  isconnected := false;
  online := false;
  if hostservuser <> nil then hostservuser.destroy;
  hostservuser := nil;

end;

procedure startup;
var
  a:integer;
  timer : tltimer;
  t : textfile;
  tr : treadtext;
  s : string;
begin
  sock := twsocket.create(nil);
  sock.onsessionconnected := sc.connected;
  sock.ondataavailable := sc.receive;
  sock.onsessionclosed := sc.closed;
  timer := tltimer.create(nil);
  timer.interval := 1000;
  timer.ontimer := sc.timehandler;
  {startup core}
  me := adduser;
  setnickname(me,opt.myservername);
  me.fullname := opt.info;
  addserver(me,opt.myservernum and 4095,maxbot,nil);
  mynum := p10inttostr(opt.myservernum and 4095,SSlen);

  {open signal port}
  {$ifndef nosignal}
  debugout('about to call signalstart');
  signalstart;
  debugout('signalstart complete');
  {$endif}
  {
  on startup stuff.
  call onstartup of any service units which i want to have running now
  }
  if not fileexists(botstartfile) then begin
    // if there is not a startbots.ini we build one with the same bots as a
    // default 2.1.x build
    assign(t,botstartfile);
    rewrite(t);
    writeln(t,'auth');
    writeln(t,'bworld');
    closefile(t);
  end;

  {for a := maxserviceunit downto 0 do begin
    if serviceunitarray[a] <> nil then
    if @serviceunitarray[a].onstartup <> nil then
    serviceunitarray[a].onstartup(serviceunitarray[a].tag);
  end;}
  assert(readtext_init(tr,botstartfile),'starting bots','can''t open '+botstartfile );
  debugout('opened '+botstartfile);

  while not readtext_eof(tr) do begin
    s := ircupper(readtext_line(tr));
    debugout('read line '+s);
    for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
      if ircupper(serviceunitarray[a].name)=s then
        if @serviceunitarray[a].onstartup <> nil then
          serviceunitarray[a].onstartup(serviceunitarray[a].tag);
  end;
  readtext_close(tr);
  debugout('finished reading '+botstartfile);
end;

procedure saveall;
var
  a : integer;
begin
  bconfig.writecfg;
  for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
   if @serviceunitarray[a].onsave <> nil then
    serviceunitarray[a].onsave(serviceunitarray[a].tag,true);
end;

procedure tsc.timehandler;
var
  a:integer;
  s:tserviceunit;
begin
  debugout('timehandler called');
  //debugout(inttostr(ord(signalsource.connected)));
  if reconnectcountdown > 0 then dec(reconnectcountdown);
  if reconnectcountdown <= 0 then begin
    if wanttoconnect then startconnect;
    reconnectcountdown := 4;
  end;

  {call timehandler of a few other units which are not bots}
  btime.timehandler;

  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then if @s.ontimer <> nil then begin
      s.ontimer(s.tag);
    end;
  end;

  //autosave
  if opt.autosave <> 0 then begin
    if (irctime mod opt.autosave)=0 then begin
      if opt.autosavenotice then send(me.idstr+' '+tok_notice+' '+opt.operchannel+' :autosave');
      saveall;

    end;
  end;

  sendflush;
end;

{
start up bots button

  for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
  if @serviceunitarray[a].onstartup <> nil then
  serviceunitarray[a].onstartup;

shutdown bots button

  for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
  if @serviceunitarray[a].onshutdown <> nil then
  serviceunitarray[a].onshutdown;

want connect button

  startconnect;

want disconnect button

  wanttoconnect := false;
  sock.close;

on terminate, call shutdown

}

procedure initapplication;
begin
  startup;
  startconnect;
end;


procedure runapplication;
begin
  messageloop
end;

procedure cleanupapplication;
begin
  globalshutdown;
end;


procedure filelog(filename,line:string);
var
  t:textfile;
begin
  assignfile(t,filename);
  {$i-}append(t);{$i+}
  if ioresult <> 0 then {$i-}rewrite(t);{$i+}
  if ioresult <> 0 then exit;
  writeln(t,'['+inttostr(unixtime)+' '+timestrshort(unixtime)+'] '+line);
  closefile(t);
end;

procedure requestsignonts(us:tuser);
begin
  if us.signonts = 0 then begin
    us.signonts := 1;
    sendparse(me.idstr+' '+TOK_WHOIS+' '+tuser(us.server.us).idstr+' '+us.name);
  end;
end;

end.
